import traceback
from nozyio.config_utils import config
import sys
from nozyio.code_to_graph import json_dict_to_code
from nozyio.utils import is_serializable
from nozyio.websocket_manager import websocket_manager

def get_handle_uid(io_type, node_id, handle_id):
    return 'node_' + node_id + '_' + io_type + '_' + handle_id

def traverse_ast_tree_and_replace(ast_node, handle_connection_map, node_id, values):
    if not isinstance(ast_node, dict):
        return ast_node
    io_type = ast_node.get('io_type')
    if io_type == 'output':
        var_id = get_handle_uid('output', node_id, ast_node.get('id'))
        return {
            "node_type": "Name",
            "id": var_id,
            "ctx": {
                "node_type": "Store"
            }
        }
    elif io_type == 'input':
        var_id = get_handle_uid('input', node_id, ast_node.get('id'))
        if var_id in handle_connection_map:
            return {
                "node_type": "Name",
                "id": handle_connection_map[var_id],
                "ctx": {
                    "node_type": "Load"
                }
            }
        elif var_id in values and values[var_id] is not None:
            return {
                "node_type": "Constant",
                "value": values[var_id],
                "kind": None
            }
        elif ast_node.get('optional'):
            print('🔥 ast_node.get("optional")', ast_node.get('optional'))
            return {
                "node_type": "Constant",
                "value": None,
                "kind": None
            }
        else:
            raise Exception(f'❌No value provided for required input "{ast_node.get("name") or ast_node.get("id") }" in node #{node_id}. Please provide a value for this input.')

    # Iterate over all fields in the node
    for key, value in ast_node.items():
        if isinstance(value, dict):
            if replace_value := traverse_ast_tree_and_replace(value, handle_connection_map, node_id, values):
                ast_node[key] = replace_value
        elif isinstance(value, list):
            # If the field is a list, check each item
            for index, item in enumerate(value):
                if isinstance(item, dict):
                    if replace_value := traverse_ast_tree_and_replace(item, handle_connection_map, node_id, values):
                        value[index] = replace_value
    
    return ast_node

def send_job_status(message: dict):
    websocket_manager.send_message_sync({
        'type': 'job_status',
        'data': message
    })

def graph_to_code(graph: dict):
    order, handle_connection_map = get_execution_order(graph)
    ast_nodes_aggr = []
    for node in order:
        ast_nodes = transform_ast_nodes(node, handle_connection_map, graph.get('values', {}))
        ast_nodes_aggr.extend(ast_nodes)
    code = json_dict_to_code({
        "node_type": "Module",
        "body": ast_nodes_aggr,
        "type_ignores": []
    })
    print('💻', code)
    return code

def get_execution_order(graph: dict):
    handle_connection_map = {}
    node_connection_map = {}
    nodes_dict = {}
    nodes = graph['nodes']
    for node in nodes:
        nodes_dict[node['id']] = node
    for edge in graph['edges']:
        # fill in handle connection map
        out_var_id = get_handle_uid('output', edge['source'], edge['sourceHandle'])
        in_var_id = get_handle_uid('input', edge['target'], edge['targetHandle'])
        handle_connection_map[out_var_id] = out_var_id
        handle_connection_map[in_var_id] = out_var_id
        # fill in node connection map
        if edge['target'] not in node_connection_map:
            node_connection_map[edge['target']] = []
        node_connection_map[edge['target']].append(nodes_dict[edge['source']])

    order = topological_sort(nodes, node_connection_map, set())
    return order, handle_connection_map

def execute_graph(graph: dict):
    order, handle_connection_map = get_execution_order(graph)
    local_vars = {}
    job_nodes = {}
    graph['job_status'] = {'status': 'RUNNING', 'nodes': job_nodes}
    try:
        for node in order:
            job_nodes[node['id']] = {'status': 'RUNNING', 'results': []}
            send_job_status(graph['job_status'])
            execute_node(node, handle_connection_map, local_vars, graph.get('values', {}))
            # after execution, get the outputs
            outputs = node.get('data', {}).get('output', [])
            for output in outputs:
                var_id = get_handle_uid('output', node['id'], output.get('id'))
                res = local_vars.get(var_id, None)
                print(f'👉{var_id}=', res)
                try:
                    job_nodes[node['id']]['results'].append(res if is_serializable(res) else f'<{type(res).__name__}>')
                except:
                    job_nodes[node['id']]['results'].append('<object>')
            job_nodes[node['id']]['status'] = 'SUCCESS'
            send_job_status(graph['job_status'])
        graph['job_status']['status'] = 'SUCCESS'
    except Exception as e:
        print('❌ error executing job', e)
        print(traceback.format_exc())
        job_nodes[node['id']]['status'] = 'FAIL'
        # job_nodes[node['id']]['error'] = str(e) + '\n' + traceback.format_exc()
        job_nodes[node['id']]['error'] = str(e)
        graph['job_status']['status'] = 'FAIL'
        send_job_status(graph['job_status'])

def topological_sort(nodes, node_connection_map:dict, visited:set):
    def dfs(node, node_connection_map:dict, visited:set, order:list):
        if node['id'] in visited:
            return
        visited.add(node['id'])
        for source_node in node_connection_map.get(node['id'], []):
            dfs(source_node, node_connection_map, visited, order)
        order.append(node)
    order = []
    for node in nodes:
        dfs(node, node_connection_map, visited, order)
    return order

def transform_ast_nodes(node, handle_connection_map, values) -> list:
    node_data = node.get('data', {})
    if not node_data.get('module') or not node_data.get('name'):
        return
    ast_nodes = node.get('data', {}).get('astNodes', [])
    for ast_node in ast_nodes:
        # traverse ast tree and replace the input_change_map with the actual value
        traverse_ast_tree_and_replace(ast_node, handle_connection_map, node['id'], values)
    return ast_nodes

def execute_node(node, handle_connection_map, local_vars, values):
    ast_nodes = transform_ast_nodes(node, handle_connection_map, values)

    code = json_dict_to_code({
        "node_type": "Module",
        "body": ast_nodes,
        "type_ignores": []
    })
    print('executing 💻', code)
    exec(code,  globals(), local_vars)
        
    return local_vars


if __name__ == '__main__':
    sys.path.append(config['package_path'])
    test_graph = {}
    execute_graph(test_graph)


import json
import os
import sys

def get_config() -> dict:
    current_dir = os.getcwd()
    default_config = {
        'package_path': os.path.join(current_dir, 'nozy_packages'),
        'model_path': 'models',
        'workflow_path': 'workflows',
        'output_path': 'outputs',
    }
    if os.path.exists("config.json"):
        with open("config.json", "r") as f:
            config = json.load(f)
    if config.get('package_path') is not None:
        config['package_path'] = os.path.join(current_dir, config['package_path'])
    
    config = {
        **default_config,
        **config,
    }
    sys.path.append(config['package_path'])
    return config

def get_root_dir():
    return os.getcwd()

config = get_config()


import ast

def infer_type_from_annotation(annotation):
    """Infer type from annotation."""
    if annotation:
        return ast.unparse(annotation)
    return "Any"

def infer_type_from_default(default_value):
    """Infer the type of a parameter from its default value if it's a primitive type."""
    if isinstance(default_value, int):
        return 'int'
    elif isinstance(default_value, str):
        return 'str'
    elif isinstance(default_value, bool):
        return 'bool'
    elif isinstance(default_value, float):
        return 'float'
    else:
        return None

def extract_function_info_from_ast(node, module_name):
    """Extract function information from an AST node."""
    function_info = {
        "name": node.name,
        "parameters": [],
        "return_type": infer_type_from_annotation(node.returns)
    }

    # Extract parameters
    for index, arg in enumerate(node.args.args):
        param_info = {
            "id": str(index),
            "name": arg.arg,
            "type": infer_type_from_annotation(arg.annotation),
            "default": None
        }
        function_info["parameters"].append(param_info)

    # Extract default values
    if node.args.defaults:
        for i, default in enumerate(node.args.defaults):
            param_info = function_info["parameters"][-(i + 1)]
            param_info["default"] = ast.unparse(default)
            inferred_type = infer_type_from_default(eval(param_info["default"]))
            if inferred_type:
                param_info["type"] = inferred_type
    astNode = {
            "node_type": "Assign",
            "value": {
                "node_type": "Call",
                "func": {
                    "node_type": "Name",
                    "id": node.name,
                    "ctx": {
                        "node_type": "Load"
                    }
                },
                "args": function_info["parameters"],
                "keywords": []
            },
            "targets": [{
                "id": "0",
                "io_type": "output",
                "type": function_info["return_type"]
            }]
        } 
    return {
        "type": "function",
        "name": node.name,
        "module": module_name,
        "input": function_info["parameters"],
        "output": [{
            "id": "0",
            "type": function_info["return_type"]
        }],
        "astNodes": [astNode]
    }

def parse_python_file(file_path, module_name):
    """Parse a Python file and extract function information."""
    with open(file_path, "r") as file:
        source_code = file.read()

    tree = ast.parse(source_code)
    functions = []

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            func_info = extract_function_info_from_ast(node, module_name)
            functions.append(func_info)

    return functions

if __name__ == "__main__":
    # Example usage
    file_path = "path/to/your/module.py"
    module_name = "module"
    functions_info = parse_python_file(file_path, module_name)
    for func in functions_info:
        print(func)


import ast
import json
import astpretty
from typing import Union


def code_to_json(code: str) -> Union[dict, list]:
    # Parse the Python code to create the AST
    class ASTToJSON(ast.NodeVisitor):
        def generic_visit(self, node):
            if isinstance(node, ast.AST):
                fields = {field: self.generic_visit(getattr(node, field)) for field in node._fields}
                return {"node_type": node.__class__.__name__, **fields}
            elif isinstance(node, list):
                return [self.generic_visit(item) for item in node]
            else:
                return node
    tree = ast.parse(code)
    astpretty.pprint(tree, indent='    ', show_offsets=False)
    # Convert the AST to JSON-compatible format
    return ASTToJSON().generic_visit(tree)



class JSONToAST:
    def convert(self, node):
        # If it's a dictionary (representing an AST node), reconstruct the node
        if isinstance(node, dict) and 'node_type' in node:
            node_type = node.pop('node_type')
            # Find the corresponding AST class
            ast_class = getattr(ast, node_type)
            # Reconstruct the AST node
            ast_node = ast_class()
            for field, value in node.items():
                setattr(ast_node, field, self.convert(value))
            # Ensure 'lineno' and 'col_offset' are set for nodes that require them
            if isinstance(ast_node, ast.AST):
                if not hasattr(ast_node, 'lineno'):
                    ast_node.lineno = 0  # Default value for missing line number
                if not hasattr(ast_node, 'col_offset'):
                    ast_node.col_offset = 0  # Default value for missing column offset
            return ast_node
        # If it's a list, recursively convert each item
        elif isinstance(node, list):
            return [self.convert(item) for item in node]
        # For everything else (constants like numbers, strings), return as-is
        else:
            return node

def json_dict_to_code(json_dict: dict) -> str:
    ast_node = JSONToAST().convert(json_dict)
    return ast.unparse(ast_node)

def json_to_ast(json_data):
    try:
        # Parse the JSON string to get a dictionary
        ast_dict = json.loads(json_data)
        # Convert the dictionary back to an AST
        return JSONToAST().convert(ast_dict)
    except Exception as e:
        print(f"Error while converting JSON to AST: {e}")
        return None

def code_to_graph(abs_path: str):
    with open(abs_path, "r") as file:
        code = file.read()
    tree = code_to_json(code)
    return tree

if __name__ == "__main__":
    code = """
from core.abc.ds import myfunction

"""
    print('python to ast json...')
    json_dict = code_to_json(code)
    json_string = json.dumps(json_dict, indent=2)
    print('➡️json graph of ast tree...', json_string)
    ast_node = json_to_ast(json_string)
    print('➡️ json back to ast tree...')
    astpretty.pprint(ast_node, indent='    ', show_offsets=False)
    print('⭐️ ast back to code...')
    code = ast.unparse(ast_node)
    
    print(code)


import math
import os
import importlib
import inspect
import fnmatch
import time
import json
import traceback
from nozyio.config_utils import config, get_root_dir
from nozyio.utils import is_serializable

def matches_blacklist(filepath, blacklist):
    """Check if the file path matches any pattern in the blacklist."""
    for pattern in blacklist:
        if fnmatch.fnmatch(filepath, pattern):
            return True
    return False


def infer_type_from_default(default_value):
    """Infer the type of a parameter from its default value if it's a primitive type."""
    if isinstance(default_value, int):
        return 'int'
    elif isinstance(default_value, str):
        return 'str'
    elif isinstance(default_value, bool):
        return 'bool'
    elif isinstance(default_value, float):
        return 'float'
    else:
        return None

def get_function_return_type(obj):
    if getattr(obj, 'NOZY_NODE_DEF', None):
        outputs = obj.NOZY_NODE_DEF.get('output', obj.NOZY_NODE_DEF.get('outputs', None))
        if outputs:
            node_outputs = [
                {
                    "id": str(index),
                    "io_type": "output",
                    "type": output if isinstance(output, str) else None,
                    **(output if isinstance(output, dict) else {})
                }
                for index, output in enumerate(outputs)
            ]
            return node_outputs

    signature = inspect.signature(obj)
    anno_type = getattr(signature.return_annotation, '__name__', str(signature.return_annotation)) if signature.return_annotation != inspect.Signature.empty else "any"
    return [{
        "id": "0",
        "io_type": "output",
        "type": anno_type
    }]


def extract_function_details(obj, module_name):
    if inspect.isfunction(obj):
        # Get function signature
        signature = inspect.signature(obj)
        ast_args = []
        nozy_node_def = getattr(obj, 'NOZY_NODE_DEF', {})
        extra_input_info = nozy_node_def.get('input', nozy_node_def.get('inputs', {}))
        
        # Collect information about parameters
        for index, (param_name, param) in enumerate(signature.parameters.items()):
            # print(f'🔥Function {obj.__name__} {param_name} {param.default} {param.annotation}')
            default = param.default if is_serializable(param.default) else None
            type = infer_type_from_default(default)
            ast_args.append({
                "id": str(param_name), 
                "io_type": "input",
                "default": default,
                "optional": param.default is not inspect.Parameter.empty,
                "name": param_name,
                "type": type if type else getattr(param.annotation, '__name__', str(param.annotation)) if param.annotation != inspect.Parameter.empty else "any",
                **extra_input_info.get(param_name, {})
            }) 
        outputs = get_function_return_type(obj)
        
        # Create the import AST node
        import_ast_node = {
            "node_type": "ImportFrom",
            "module": module_name,
            "names": [{
                "node_type": "alias",
                "name": obj.__name__,
                "asname": None
            }],
            "level": 0
        }
        
        # Create the assign AST node
        assign_ast_node = {
            "node_type": "Assign",
            "value": {
                "node_type": "Call",
                "func": {
                    "node_type": "Name",
                    "id": obj.__name__,
                    "ctx": {
                        "node_type": "Load"
                    }
                },
                "args": ast_args,
                "keywords": []
            },
            "targets": outputs if len(outputs) == 1 else [{
                "node_type": "Tuple",
                "elts": outputs,
                "ctx": {
                    "node_type": "Store"
                }
            }]
        } 
        
        return {
            "type": "function",
            "name": obj.__name__,
            "module": module_name,
            "input": ast_args,
            "output": outputs,
            "astNodes": [import_ast_node, assign_ast_node]
        }
def extract_class_details(cls_obj, module_name):
    """Extract detailed information about a class, including its methods."""
    class_name = cls_obj.__name__

    # List methods of the class
    methods = []
    for name, method in inspect.getmembers(cls_obj, predicate=inspect.isfunction):
        # Only include methods defined in the class itself
        if method.__qualname__.startswith(cls_obj.__name__):
            method_info = extract_function_details(method, module_name)
            methods.append(method_info)
    
    return {
        "type": "class",
        "name": class_name,
        "module": module_name,
        "methods": methods
    }

package_path = config['package_path']
def list_functions_classes(py_file_path):
    if not os.path.isabs(py_file_path):
        raise ValueError(f"Please provide an absolute path, '{py_file_path}' is not absolute.")
    print('py_file_path', py_file_path, 'package_path', package_path)
    if os.path.commonpath([py_file_path, package_path]) == package_path:
        package_rel_path = os.path.relpath(py_file_path, package_path)
    else:
        root_dir = get_root_dir()
        package_rel_path = os.path.relpath(py_file_path, root_dir)
    # Get the module name by replacing path separators with dots and removing the file extension
    module_name = package_rel_path.replace(os.sep, '.').rsplit('.', 1)[0]
    print('👾 module_name', module_name)
    # load and execute the module to inspect its contents
    module = importlib.import_module(module_name)

    details = []
    for name, obj in inspect.getmembers(module):
        if inspect.isfunction(obj) and obj.__module__ == module.__name__:
            # Handle functions
            print(f'😁 find function {name}')
            func_info = extract_function_details(obj, module.__name__)
            details.append(func_info)
        # elif inspect.isclass(obj) and obj.__module__ == module.__name__:
        #     # Handle classes
        #     class_info = extract_class_details(obj, module.__name__)
        #     details.append(class_info)
    
    return details

def refresh_node_def(graph: dict):
    nodes = graph.get('nodes')
    for node in nodes:
        try:
            module_name = node.get('data', {}).get('module')
            name = node.get('data', {}).get('name')
            if module_name and name:
                # print(f'🔥 refresh node def for {module_name} {name}')
                module = importlib.import_module(module_name)
                obj = getattr(module, name)
                node_def = extract_function_details(obj, module_name)
                if isinstance(node_def, dict):
                    node['data'] = node_def
        except Exception as e:
            print(f'❌ error refreshing node def for {module_name} {name}: {e}')
            print(traceback.format_exc())
            node['data']['import_error'] = str(e)
    return graph

def scan_directory(directory):
    print('🗂️ finding python files in', directory)
    if not os.path.isdir(directory):
        raise ValueError(f"Please provide an absolute path, '{directory}' is not absolute.")
    
    children = []
    root_dir = get_root_dir()
    # Use os.listdir to avoid recursion
    for file in os.listdir(directory):
        full_path = os.path.join(directory, file)
        if file.endswith('.py'):
            print(f'📄 find python file {file}')
            functions = list_functions_classes(full_path)
            children.append({
                "type": "file",
                "name": file,
                "path": os.path.relpath(full_path, root_dir),
                "functions": functions
            })
        if os.path.isdir(full_path):
            children.append({
                "type": "folder",
                "path": os.path.relpath(full_path, root_dir),
                "name": file,
            })
    
    return children

project_root = os.path.dirname(os.path.dirname(__file__))
def scan_directories(directories, blacklist=None):
    if blacklist is None:
        blacklist = []
        
    all_modules_info = {}
    for path in directories:
        # path can be relative or absolute
        abs_path = os.path.join(project_root, path)
        print(f"🗂️ Scanning directory: {abs_path}")
        modules_info = scan_directory(abs_path, blacklist=blacklist)
        all_modules_info.update(modules_info)
    return all_modules_info

if __name__ == "__main__":
    print("Scanning modules...")
    base_directories = ["web"] 
    blacklist = ["*/install.py"]  # Add more patterns as needed
    start_time = time.time()
    modules_info = scan_directories(base_directories, blacklist=blacklist)
    print(json.dumps(modules_info, indent=2))
    # print(f"Scanning complete in {time.time() - start_time:.2f} seconds")


import asyncio
from aiohttp.web import WebSocketResponse
from typing import Dict

class WebSocketManager:
    def __init__(self):
        self.connections: Dict[str, WebSocketResponse] = {}  # Dictionary to store user_id: websocket pairs

    async def set_connection(self, ws, user_id='default'):
        self.connections[user_id] = ws

    async def send_message(self, message, user_id='default'):
        ws = self.connections.get(user_id)
        if ws is None or ws.closed:
            print(f"WebSocket for user {user_id} is not connected or is closed.")
            return
        await ws.send_json(message)
        # print(f"Sent message to user {user_id}: {message}")

    def send_message_sync(self, message, user_id='default'):
        future = asyncio.run_coroutine_threadsafe(
            self.send_message(message, user_id),
            self.loop  # Use the stored loop
        )

# Create a global instance of the WebSocketManager
websocket_manager = WebSocketManager()


import asyncio
import json
import sys
from aiohttp import web
import os
from nozyio.code_to_graph import code_to_graph
from nozyio.ast_execution import execute_graph, graph_to_code
from nozyio.scan_modules import refresh_node_def, scan_directory
from nozyio.config_utils import config, get_root_dir
from nozyio.websocket_manager import websocket_manager
from nozyio.job_queue_manager import on_cleanup, process_queue, run_in_executor

WEB_PATH = os.path.join(os.path.dirname(__file__), 'web/dist')

def endpoint(path, method='GET'):
    def decorator(func):
        func.route_info = {'path': path, 'methods': [method]}
        return func
    return decorator

@endpoint('/queue_status')
async def handle_queue_status(request):
    task_queue = request.app['task_queue']
    return web.json_response({
        'queue_size': task_queue.qsize(),
        'tasks_pending': not task_queue.empty()
    })

@endpoint('/code_to_graph')
async def handle_code_to_graph(request):
    rel_path = request.rel_url.query['path']
    abs_path = os.path.join(config['package_path'], rel_path)
    return web.json_response(code_to_graph(abs_path))

@endpoint('/queue_job', method='POST')
async def handle_queue_job(request):
    body = await request.json()
    task_queue = request.app['task_queue']
    await task_queue.put(body)  # Add the task to the queue
    print('Job queued')
    return web.json_response({'status': 'queued'})

@endpoint('/run_job', method='POST')
async def handle_run_job(request):
    body = await request.json()
    await run_in_executor(execute_graph, body)
    return web.json_response({'status': 'ok'})

@endpoint('/list_package_children')
async def handle_scan_directories(request):
    rel_path = request.rel_url.query['path']
    blacklist = config.get('blacklist', [])
    abs_path = os.path.join(get_root_dir(), rel_path)
    children = scan_directory(abs_path)
    return web.json_response(children)

@endpoint('/graph_to_code', method='POST')
async def handle_graph_to_code(request):
    graph = await request.json()
    code = graph_to_code(graph)
    return web.json_response(code)

@endpoint('/ws')
async def websocket_handler(request):
    ws = web.WebSocketResponse()
    await ws.prepare(request)

    print('WebSocket connection established')
    await websocket_manager.set_connection(ws)

    try:
        async for msg in ws:
            if msg.type == web.WSMsgType.TEXT:
                # Handle incoming messages if needed
                print(f"Received message: {msg.data}")
                await websocket_manager.send_message(f"Echo: {msg.data}")
            elif msg.type == web.WSMsgType.ERROR:
                print(f"WebSocket connection closed with exception {ws.exception()}")

    finally:
        print('WebSocket connection closed')

    return ws

@endpoint('/refresh_node_def', method='POST')
async def handle_refresh_node_def(request):
    graph = await request.json()
    new_graph = refresh_node_def(graph)
    return web.json_response(new_graph)

@endpoint('/')
async def handle_index(request):
    return web.FileResponse(os.path.join(WEB_PATH, 'index.html'))

# Middleware to handle CORS
@web.middleware
async def cors_middleware(request, handler):
    if request.method == 'OPTIONS':
        # Handle preflight CORS requests
        response = web.Response(status=204)
    else:
        response = await handler(request)
    response.headers['Access-Control-Allow-Origin'] = '*'
    return response

async def on_startup(app):
    app['task_queue'] = asyncio.Queue()
    # Start the background task to process the queue
    app['process_queue_task'] = asyncio.create_task(process_queue(app))
    websocket_manager.loop = asyncio.get_running_loop()
    
def start_server():
    allow_cors = '--allow-cors' in sys.argv
    if allow_cors:
        print('🧪allow cors=', allow_cors)
    app = web.Application(middlewares=[cors_middleware] if allow_cors else [])
    
    # Register routes using the endpoint decorator
    for name, func in globals().items():
        if callable(func) and hasattr(func, 'route_info'):
            route_info = func.route_info
            for method in route_info['methods']:
                app.router.add_route(method, route_info['path'], func)
    
    if os.path.exists(WEB_PATH):
        app.router.add_static('/', WEB_PATH, show_index=True)
    
    app.on_startup.append(on_startup)
    app.on_cleanup.append(on_cleanup)

    # Check for --listen argument
    if '--listen' in sys.argv:
        host = '0.0.0.0'
    else:
        host = '127.0.0.1'

    web.run_app(app, host=host, port=7070)

if __name__ == "__main__":
    print("Starting aiohttp server...")
    start_server()



import math


def is_serializable(value):
    """Check if a value is JSON serializable."""
    if isinstance(value, float):
        if math.isinf(value) or math.isnan(value):
            return False
        return True
    elif isinstance(value, (int, str, bool)):
        return True
    elif isinstance(value, (list, tuple)):
        return all(is_serializable(item) for item in value)
    elif isinstance(value, dict):
        return all(isinstance(key, str) and is_serializable(val) for key, val in value.items())
    return False
    
    


import asyncio
from nozyio.ast_execution import execute_graph
import concurrent.futures

executor = concurrent.futures.ThreadPoolExecutor(max_workers=6)  # Adjust the number of workers as needed

async def run_in_executor(func, *args):
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(executor, func, *args)

async def process_queue(app):
    print('inside process_queue')
    task_queue = app['task_queue']
    while True:
        task = await task_queue.get()
        try:
            await run_in_executor(execute_graph, task)
        finally:
            task_queue.task_done()


async def on_cleanup(app):
    # Cancel the process_queue task
    process_queue_task = app.get('process_queue_task')
    if process_queue_task:
        process_queue_task.cancel()
        try:
            await process_queue_task
        except asyncio.CancelledError:
            pass
